#!/usr/bin/q --
#! -c main ARGS

/* qcwrap: Create a C wrapper program for running a Q script.
   $Id: qcwrap.q,v 1.3 2004/04/14 08:38:16 agraef Exp $ */

/* This file is part of the Q programming system.

   The Q programming system is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   The Q programming system is distributed in the hope that it will be
   useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/* USAGE: To bootstrap this script (i.e., create a wrapper for qcwrap itself),
   run the following command (Linux/Unix):

   $ ./qcwrap.q qcwrap.q

   On Windows, or other systems which don't recognize `#!', you'll have to
   explicitly invoke the script with the Q interpreter:

   $ q qcwrap.q qcwrap.q

   Be patient, generating the C source for the bytecode of the script takes a
   while... If all went well, you should now have a qcwrap.c file. Compile the
   C program as follows (taking gcc as an example):

   $ gcc -o qcwrap qcwrap.c -lqint

   Voila! You should now have a qcwrap executable which can be executed
   instead of the original qcwrap.q script. To test it:

   $ mv qcwrap.c qcwrap.c.bak
   $ ./qcwrap qcwrap.q
   $ diff qcwrap.c.bak qcwrap.c

   The two files should not differ.

   To install the qcwrap executable, just copy it to some directory on your
   PATH. Whenever you need to create a C wrapper for a Q script, say `foo.q',
   just run `qcwrap foo.q' and compile the resulting C file foo.c with your C
   compiler (don't forget to link in libqint). */

private qcwrap, qc, generate, header, code, body;
private exception, cleanup;

basename EXT NAME = sub NAME 0 (#NAME-#EXT-1)
		      if sub NAME (#NAME-#EXT) (#NAME-1) = EXT;
		  = NAME otherwise;

exitstatus STATUS = STATUS; // for Windows compatibility

/* The main program. Make sure to check for exceptions (e.g., Ctl-C) and
   perform necessary cleanup. */

main ARGS

= catch (exception ARGS) (qcwrap ARGS);

exception ARGS _

= catch (exception ARGS)
  (cleanup ARGS || fprintf ERROR "%s: compilation aborted\n" (ARGS!0) ||
   exit 1);

cleanup ARGS

= unlink QCODE || unlink CFILE
    if isfile (fopen QFILE "r")
    where QFILE = ARGS!1, QCODE = basename ".q" QFILE++".out",
      CFILE = basename ".q" QFILE++".c";

qcwrap ARGS

= printf "Usage: %s script-name\n" (ARGS!0) || exit 1 if #ARGS<>2;

= fprintf ERROR "%s: file `%s' not found\n" (ARGS!0,ARGS!1) || exit 1
    if not isfile (fopen (ARGS!1) "r");

= qc QFILE QCODE || generate (fopen CFILE "w") QFILE CFILE
  (bread (fopen QCODE "rb") (st_size (stat QCODE))) ||
  unlink QCODE || exit 0
    where QFILE = ARGS!1, QCODE = basename ".q" QFILE++".out",
      CFILE = basename ".q" QFILE++".c";

/* Invoke qc, check its exit code. */

qc QFILE QCODE

= fprintf ERROR "%s: qc returned exit code %d\n" (ARGS!0,RET) ||
  exit 1 if RET <> 0
    where CMD = sprintf "qc -o %s %s" (str QCODE,str QFILE),
      RET = exitstatus (system CMD);

= () otherwise;

/* Parse options from the source file. */

parseopt S

= (OPTS,"")
    where [OPTS] = regex "" "^(-[hVnvwdeiq]+)$" S (reg 1);

= (OPTS,ARG)
    where [(OPTS,ARG)] =
      regex "" "^(-[hVnvwdeiq]*[lopstc])[ \t]*(.*)$" S (reg 1,reg 2);

= (OPTS,ARG)
    where [(OPTS,ARG)] = regex "" "^(--[^=]+)=[ \t]*(.*)$" S (reg 1,reg 2);

= (OPTS,"")
    where [OPTS] = regex "" "^(--.+)$" S (reg 1);

= ("--","") if S = "--";

= () otherwise;

shebang S

= parseopt S
    where [S] = regex "" "^#![ \t]*[^ \t]+[ \t]+(.*)$" S (reg 1);

= () otherwise;

optline S

= parseopt S
    where [S] = regex "" "^#![ \t]+(.*)$" S (reg 1);

= () otherwise;

opts OPTS F

= opts [(OPT,ARG)|OPTS] F where (OPT,ARG) = optline (freads F);

= reverse OPTS otherwise;

getopts F

= opts [(OPT,ARG)] F where (OPT,ARG) = shebang (freads F);

= opts [] F otherwise;

/* Generate the C program. */

generate F QFILE CFILE CODE

= header F QFILE CFILE || code F 0 CODE || body F QFILE
    if isbytestr CODE;

= fprintf ERROR "%s: couldn't find the byte code file\n" (ARGS!0) || exit 1
    otherwise;

/* Generate the header of the program. */

header F QFILE CFILE

= fprintf F "/* %s: generated automatically by qcwrap from %s */\n"
  (CFILE,QFILE) ||
  fprintf F "\n#include <qint.h>\n\n" ();

/* Generate the body of the program. If there is a shebang like, say,
   `#!/usr/bin/q -c EXPR' at the beginning of the source script, use EXPR as
   the expression to be evaluated. Also evaluate any expressions given on
   subsequent -c option lines. Note that the argument *must* be an expression,
   other interpreter commands like `def' etc. will not work. */

copt (OPT,_)

= (#OPT > 2) and then (OPT = sub "--command" 0 (#OPT-1))
    if sub OPT 0 1 = "--";

= pos "c" OPT >= 0 otherwise;

body F QFILE

= fwrites F "\n\
int main(int argc, char *argv[])\n\
{\n\
  int status;\n\
  status = qexecvx(code, sizeof(code), argc, argv);\n\
" ||
  do (fprintf F "\
  if (status == 0) qeval(%s, &status);\n\
".str) CMDS ||
  fwrites F "\
  exit(status);\n\
}\n\
"
    where CMDS = map snd (filter copt (getopts (fopen QFILE "r")));

/* Encode the bytecode of the script as an unsigned char[]. */

code F I CODE

= fprintf F "\n};\n" () if I >= #CODE;

= fprintf F "unsigned char code[] = {\n  0x%02x" (CODE!0) ||
  code F 1 CODE if I = 0;

= fprintf F ",\n  0x%02x" (CODE!I) || code F (I+1) CODE if I mod 15 = 0;

= fprintf F ",0x%02x" (CODE!I) || code F (I+1) CODE otherwise;
